//
// Copyright 2004 Andras Varga
//
// This library is free software, you can redistribute it and/or modify
// it under  the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation;
// either version 2 of the License, or any later version.
// The library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.
//


#include "DNITCPSessionApp.h"
#include <sstream>
#include <string>
#include <iostream>
#include <algorithm>
#include <iterator>
#include <vector>
#include <stdlib.h>

#include "ByteArrayMessage.h"
#include "GenericAppMsg_m.h"
#include "IPvXAddressResolver.h"


Define_Module(DNITCPSessionApp);

simsignal_t DNITCPSessionApp::rcvdPkSignal = SIMSIGNAL_NULL;
simsignal_t DNITCPSessionApp::sentPkSignal = SIMSIGNAL_NULL;
simsignal_t DNITCPSessionApp::recvIndicationsSignal = SIMSIGNAL_NULL;


std::vector<std::string> &split(const std::string &s, char delim, std::vector<std::string> &elems) {
    std::stringstream ss(s);
    std::string item;
    while (std::getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}


std::vector<std::string> split(const std::string &s, char delim) {
    std::vector<std::string> elems;
    split(s, delim, elems);
    return elems;
}
//usage: std::vector<std::string> x = split("one:two::three", ':');


//////////////////////////////////////////////////////////////////////////////////////////////////////


bool isValidSegement(const char *script){
    std::string input(script);
    std::vector<std::string> tokens;

    std::istringstream iss(input);
    std::copy(std::istream_iterator<std::string>(iss),
                   std::istream_iterator<std::string>(),
                   std::back_inserter<std::vector<std::string> >(tokens));

    int bytes = 0;
    double think = 0.0;
    int numpics = 0;
    double num = 0;

    //std::cout << "Analyzing segment: " << script<< " " << '\n';

    if(tokens.size() >= 7 && tokens.at(0).compare("send")==0 && tokens.at(2).compare("every")==0 && tokens.at(4).compare("for")==0 && tokens.at(6).compare("times")==0){
        std::istringstream t2(tokens[1]);
        std::istringstream t4(tokens[3]);
        std::istringstream t6(tokens[5]);

//        std::strin s = tokens[1];
//        !s.empty() && s.find_first_not_of("0123456789.") == std::string::npos;

        char* p;
        long converted = strtod(tokens[1].c_str(), &p);
        if (*p) {
            std::cout << "Parsing unsuccessful! Token ("<< tokens[1] <<") of " << input << " is not a number" << '\n';
            return false;
        }
        converted = strtod(tokens[3].c_str(), &p);
        if (*p) {
            std::cout << "Parsing unsuccessful! Token ("<< tokens[3] <<") of " << input << " is not a number" << '\n';
            return false;
        }
        converted = strtod(tokens[5].c_str(), &p);
        if (*p) {
            std::cout << "Parsing unsuccessful! Token ("<< tokens[5] <<") of " << input << " is not a number" << '\n';
            return false;
        }


        return true;
    }
    else if(tokens.size() == 2 && tokens.at(0).compare("send")==0){
        char* p;
        long converted = strtod(tokens[1].c_str(), &p);
        if (*p) {
            std::cout << "Parsing unsuccessful! Token ("<< tokens[1] <<") of " << input << " is not a number" << '\n';
            return false;
        }

        return true;
    }
    else if(tokens.size() == 2 && tokens.at(0).compare("wait")==0){
        char* p;
        long converted = strtod(tokens[1].c_str(), &p);
        if (*p) {
            std::cout << "Parsing unsuccessful! Token ("<< tokens[1] <<") of " << input << " is not a number" << '\n';
            return false;
        }
        return true;
    }
    return false;
}


void DNITCPSessionApp::parseSpecialScriptShortSegments(const char *script){
//    const char *s = script;

    simtime_t start = 0.1;


    // UPDATE APRIL 2015
    //segment (|segment)*  - doable
    //segment = send XXXXX every YYYYY for ZZZZZ times; - doable, even done
    //segment = send XXXXX; - doable
    //segment = wait XXXXX; - ??

    std::vector<std::string> segments = split(script, ';');

    simtime_t curr = start;
//    int maxrngs = ev.getNumRNGs();
    int bytes = 0;
    double think = 0.0;
    int numpics = 0;
    int num = 0;


    //for(std::vector<T>::iterator it = segments.begin(); it != segments.end(); ++it) {
    for ( auto &segment : segments ) {

        if(isValidSegement(segment.c_str())){

            if(segment[0]== ' ')
            {
                segment.erase(segment.begin());
            }

            std::vector<std::string> tokens = split(segment.c_str(), ' ');

            std::cout << "\t Building commands out of segment " << segment << '\n';
//            std::cout << "\t tokens.size() " << tokens.size() << '\n';
//            std::cout << "\t tokens.at(0) " << tokens.at(0) << '\n';
            if(tokens.size() > 2){
//                std::cout << "\t tokens.at(2) " << tokens.at(2) << '\n';
//                std::cout << "\t tokens.at(4) " << tokens.at(4) << '\n';
//                std::cout << "\t tokens.at(6) " << tokens.at(6) << '\n';
            }

//            std::cout << "PRE_PARSING: Actual time pointer curr has value "<< curr << '\n';

            if(tokens.size() >= 7 && tokens.at(0).compare("send")==0 && tokens.at(2).compare("every")==0 && tokens.at(4).compare("for")==0 && tokens.at(6).compare("times")==0){
                //create commands for loop
//                std::cout << "Parsing loop segment " << '\n';
                bytes = 0;
                think = 0.0;
                numpics = 0;
                std::istringstream ( tokens[1] ) >> bytes;
                std::istringstream ( tokens[3] ) >> think;
                std::istringstream ( tokens[5] ) >> numpics;

//                std::cout<<"bytes_str: "<<bytes<<std::endl;
//                std::cout<<"think_str: "<<think<<std::endl;
//                std::cout<<"numpics_str: "<<numpics<<std::endl;

                for(int i = 1 ; i <= numpics ; i++){
                    Command cmd;
                    cmd.tSend = curr;
                    cmd.numBytes = bytes;
                    //int rngnum = intrand(maxrngs);

                    //simtime_t expThinkTime = exponential(think, rngnum);
                    simtime_t thinkTime = think;

                    //curr += expThinkTime;
                    curr += thinkTime;
                    commands.push_back(cmd);
                }
            }
            else if(tokens.size() == 2 && tokens.at(0).compare("send")==0){
                //create command for send
//                std::cout << "Parsing send segment " << '\n';
                bytes = 0;
                std::istringstream ( tokens[1] ) >> bytes;
                Command cmd;
                cmd.tSend = curr;
                cmd.numBytes = bytes;
                //curr += 0;
                commands.push_back(cmd);


            }
            else if(tokens.size() == 2 && tokens.at(0).compare("wait")==0){
                //create commands for wait - only if a send follows
//                std::cout << "Parsing wait segment " << '\n';
                think = 0;
                std::istringstream ( tokens[1] ) >> think;
                curr += think;
            }
        }
        else{
            std::cout << "Segment " << segment << " is invalid " << '\n';
        }
//        std::cout << "POST_PARSING: Actual time pointer curr has value "<< curr << '\n';
    }

//   for (CommandVector::iterator i = commands.begin(); i != commands.end(); ++i)
//   {
////         waitUntil(i->tSend);
//       std::cout << "Command: "<<i->tSend <<"@" << i->numBytes << " bytes" <<std::endl;
//   }

}

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

void DNITCPSessionApp::parseSpecialScriptShort(const char *script){
    const char *s = script;

      simtime_t start = 0.1;

//      std::cout<<"PARSING special short with input: "<<s<<std::endl;

      //send XXXXX every YYYYY for ZZZZZ times;

      // UPDATE APRIL 2015
      //segment (|segment)*  - doable
      //segment = send XXXXX every YYYYY for ZZZZZ times; - doable, even done
      //segment = send XXXXX; - doable
      //segment = wait XXXXX; - ??

      std::string input(script);
      std::vector<std::string> tokens;

      std::istringstream iss(input);
      std::copy(std::istream_iterator<std::string>(iss),
               std::istream_iterator<std::string>(),
               std::back_inserter<std::vector<std::string> >(tokens));

      int bytes = 0;
      double think = 0.0;
      int numpics = 0;
      if(tokens.size() >= 7 && tokens.at(0).compare("send")==0 && tokens.at(2).compare("every")==0 && tokens.at(4).compare("for")==0 && tokens.at(6).compare("times;")==0){

          std::istringstream ( tokens[1] ) >> bytes;
          std::istringstream ( tokens[3] ) >> think;
          std::istringstream ( tokens[5] ) >> numpics;
      }
      else{
          std::cout << "Parsing unsuccessful! " << input << '\n';
      }

        std::cout<<"input: "<<input<<std::endl;
        std::cout<<"bytes_str: "<<bytes<<std::endl;
        std::cout<<"think_str: "<<think<<std::endl;
        std::cout<<"numpics_str: "<<numpics<<std::endl;


        simtime_t curr = start;
        int maxrngs = ev.getNumRNGs();
        for(int i = 1 ; i <= numpics ; i++){
            Command cmd;
            cmd.tSend = curr;
            cmd.numBytes = bytes;
            int rngnum = intrand(maxrngs);


            simtime_t expThinkTime = exponential(think, rngnum);

            curr += expThinkTime;
            commands.push_back(cmd);
        }

//
//        for (CommandVector::iterator i = commands.begin(); i != commands.end(); ++i)
//       {
////         waitUntil(i->tSend);
//           //std::cout << "Command: "<<i->tSend <<"@" << i->numBytes << " bytes" <<std::endl;
//       }


//          int maxrngs = ev.getNumRNGs();
//          int rngnum = intrand(maxrngs);
//          const char *s0 = s;
//          simtime_t thinkTime = strtod(s, &const_cast<char *&>(s));
//          simtime_t expThinkTime = exponential(thinkTime, rngnum);    // now + think
//          start += expThinkTime;    // now + think
//          cmd.tSend = start;      // now+think == next trnasmission

}

void DNITCPSessionApp::parseSpecialScript(const char *script){
    const char *s = script;

    parseSpecialScriptShort(script);

      simtime_t start = 0.0;

      while (*s)
      {
          Command cmd;

          // parse time
          while (isspace(*s))
              s++;

          if (!*s || *s == ';')
              break;

          int maxrngs = ev.getNumRNGs();
          int rngnum = intrand(maxrngs);
          const char *s0 = s;
          simtime_t thinkTime = strtod(s, &const_cast<char *&>(s));
          simtime_t expThinkTime = exponential(thinkTime, rngnum);    // now + think
          start += expThinkTime;    // now + think
          //std::cout<<"Think was: "<<thinkTime<<", expThink is: "<<expThinkTime<<", now would be: "<<start<<std::endl;
          cmd.tSend = start;      // now+think == next trnasmission


          if (s == s0)
              throw cRuntimeError("Syntax error in SPECIAL script: simulation time expected");

          // parse number of bytes
          while (isspace(*s))
              s++;

          if (!isdigit(*s))
              throw cRuntimeError("Syntax error in SPECIAL script: number of bytes expected");

          cmd.numBytes = atoi(s);

          while (isdigit(*s))
              s++;

          // add command
          commands.push_back(cmd);

          // skip delimiter
          while (isspace(*s))
              s++;

          if (!*s)
              break;

          if (*s!=';')
              throw cRuntimeError("Syntax error in SPECIAL script: separator ';' missing");

          s++;

          while (isspace(*s))
              s++;
      }
}

void DNITCPSessionApp::parseScript(const char *script)
{
    const char *s = script;

    simtime_t start = 0.0;

    while (*s)
    {
        Command cmd;

        // parse time
        while (isspace(*s))
            s++;

        if (!*s || *s == ';')
            break;

        const char *s0 = s;
        cmd.tSend = strtod(s, &const_cast<char *&>(s));


        if (s == s0)
            throw cRuntimeError("Syntax error in NORMAL script: simulation time expected");

        // parse number of bytes
        while (isspace(*s))
            s++;

        if (!isdigit(*s))
            throw cRuntimeError("Syntax error in NORMAL script: number of bytes expected");

        cmd.numBytes = atoi(s);

        while (isdigit(*s))
            s++;

        // add command
        commands.push_back(cmd);

        // skip delimiter
        while (isspace(*s))
            s++;

        if (!*s)
            break;

        if (*s!=';')
            throw cRuntimeError("Syntax error in NORMAL script: separator ';' missing");

        s++;

        while (isspace(*s))
            s++;
    }
}

void DNITCPSessionApp::count(cMessage *msg)
{
    if (msg->isPacket())
    {
        if (msg->getKind() == TCP_I_DATA || msg->getKind() == TCP_I_URGENT_DATA)
        {
            packetsRcvd++;
            cPacket *packet = PK(msg);
            bytesRcvd += packet->getByteLength();
            emit(rcvdPkSignal, packet);
        }
        else
        {
            EV << "TCPSessionApp received unknown message (kind:" << msg->getKind()
               << ", name:" << msg->getName() << ")\n";
        }
    }
    else
    {
        indicationsRcvd++;
        emit(recvIndicationsSignal, msg);
    }
}

void DNITCPSessionApp::waitUntil(simtime_t t)
{
    if (simTime()>=t)
        return;

    cMessage *timeoutMsg = new cMessage("timeout");
    scheduleAt(t, timeoutMsg);
    cMessage *msg = NULL;

    while ((msg = receive()) != timeoutMsg)
    {
        count(msg);
        socket.processMessage(msg);
    }

    delete timeoutMsg;
}

cPacket* DNITCPSessionApp::genDataMsg(long sendBytes)
{
    switch (socket.getDataTransferMode())
    {
        case TCP_TRANSFER_BYTECOUNT:
        case TCP_TRANSFER_OBJECT:
        {
            cPacket *msg = NULL;
            msg = new cPacket("data1");
            msg->setByteLength(sendBytes);
            return msg;
        }

        case TCP_TRANSFER_BYTESTREAM:
        {
            ByteArrayMessage *msg = new ByteArrayMessage("data1");
            unsigned char *ptr = new unsigned char[sendBytes];

            for (int i = 0; i < sendBytes; i++)
                ptr[i] = (bytesSent + i) & 0xFF;

            msg->getByteArray().assignBuffer(ptr, sendBytes);
            msg->setByteLength(sendBytes);
            return msg;
        }

        default:
            throw cRuntimeError("Invalid TCP data transfer mode: %d", socket.getDataTransferMode());
    }
}

void DNITCPSessionApp::activity()
{
    packetsRcvd = indicationsRcvd = 0;
    bytesRcvd = bytesSent = 0;
    WATCH(packetsRcvd);
    WATCH(bytesRcvd);
    WATCH(indicationsRcvd);

    rcvdPkSignal = registerSignal("rcvdPk");
    sentPkSignal = registerSignal("sentPk");
    recvIndicationsSignal = registerSignal("recvIndications");

    // parameters
    const char *localAddress = par("localAddress");
    int localPort = par("localPort");
    const char *connectAddress = par("connectAddress");
    int connectPort = par("connectPort");

    bool active = par("active");
    simtime_t tOpen = par("tOpen");
    simtime_t tSend = par("tSend");
    long sendBytes = par("sendBytes");
    simtime_t tClose = par("tClose");

    const char *specialscript = par("specialSendScript");
    //parseSpecialScript(specialscript);
    //parseSpecialScriptShort(specialscript );
    parseSpecialScriptShortSegments(specialscript);

    const char *script = par("sendScript");
    parseScript(script);



    if (sendBytes > 0 && commands.size() > 0)
        throw cRuntimeError("Cannot use both sendScript and tSend+sendBytes");

    socket.setOutputGate(gate("tcpOut"));

    // open
    waitUntil(tOpen);

    socket.readDataTransferModePar(*this);
    socket.bind(*localAddress ? IPvXAddress(localAddress) : IPvXAddress(), localPort);

    EV << "issuing OPEN command\n";

    if (ev.isGUI())
        getDisplayString().setTagArg("t", 0, active ? "connecting" : "listening");

    if (active)
        socket.connect(IPvXAddressResolver().resolve(connectAddress), connectPort);
    else
        socket.listenOnce();

    // wait until connection gets established
    while (socket.getState() != TCPSocket::CONNECTED)
    {
        socket.processMessage(receive());

        if (socket.getState() == TCPSocket::SOCKERROR)
            return;
    }

    EV << "connection established, starting sending\n";
    if (ev.isGUI())
        getDisplayString().setTagArg("t", 0, "connected");

    // send
    if (sendBytes > 0)
    {
        waitUntil(tSend);
        EV << "sending " << sendBytes << " bytes\n";
        cPacket *msg = genDataMsg(sendBytes);
        bytesSent += sendBytes;
        emit(sentPkSignal, msg);
        socket.send(msg);
    }

    for (CommandVector::iterator i = commands.begin(); i != commands.end(); ++i)
    {
        waitUntil(i->tSend);
        EV << "sending " << i->numBytes << " bytes\n";
        cPacket *msg = genDataMsg(i->numBytes);
        bytesSent += i->numBytes;
        emit(sentPkSignal, msg);
        socket.send(msg);
    }

    // close
    if (tClose >= 0)
    {
        waitUntil(tClose);
        EV << "issuing CLOSE command\n";

        if (ev.isGUI())
            getDisplayString().setTagArg("t", 0, "closing");

        socket.close();
    }

    // wait until peer closes too and all data arrive
    for (;;)
    {
        cMessage *msg = receive();
        count(msg);
        socket.processMessage(msg);
        if (ev.isGUI())
            getDisplayString().setTagArg("t", 0, TCPSocket::stateName(socket.getState()));
    }
}

void DNITCPSessionApp::finish()
{
    EV << getFullPath() << ": received " << bytesRcvd << " bytes in " << packetsRcvd << " packets\n";
    recordScalar("bytesRcvd", bytesRcvd);
    recordScalar("bytesSent", bytesSent);
}

