package dni3toOmnetTransformation

import OMNET.NedFile
import OMNET.OMNETPackage
import OMNET.impl.OMNETPackageImpl
import OMNET.OMNETFactory
import OMNET.impl.OMNETFactoryImpl
import dnimm3.Common
import dnimm3.Dnimm3Package
import dnimm3.NamedElement
import dnimm3.NetworkInfrastructure
import dnimm3.Node
import java.io.FileOutputStream
import org.eclipse.emf.common.util.URI
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl
import OMNET.Network
import OMNET.SubModule
import OMNET.impl.SubModuleImpl
import OMNET.SimpleModule
import OMNET.impl.SimpleModuleImpl
import OMNET.LocalType
import OMNET.impl.LocalTypeImpl
import OMNET.impl.NetworkImpl
import OMNET.Channel
import OMNET.Direction
import OMNET.Connection
import OMNET.ConnectionGroup
import OMNET.Connections
import dnimm3.SDN
import OMNET.InterfaceType
import dnimm3.NetworkInterface
import OMNET.CompoundModule
import dnimm3.CommunicatingApplication
import dnimm3.Flow
import javax.swing.text.FlowView.FlowStrategy
import OMNET.OmnetppIniGroup
import dnimm3.End
import dnimm3.Link
import java.util.stream.DoubleStream.Builder
import OMNET.DoubleParameter
import dnimm3.PerformanceNetworkInterface
import dnimm3.DNIUnit
import dnimm3.DataUnit
import dnimm3.FlowRoute
import dnimm3.ConstantDoubleVariable

class Main {
	val static NAME = "dni3toomnet"
	
	val static fac = OMNETFactory.eINSTANCE
	val static omn = OMNETPackage.eINSTANCE
	
	//extension OMNETFactory = OMNETFactory.eINSTANCE
	//extension OMNETPackage = OMNETFactory.OMNETPackage

	public def static void main(String[] args) {
		
		var fname = "scenario0B_1.dnimm3"
		if (args.length < 1) {
			println(NAME +  " expects a filename as its first input, see --help for more information. Continuing with default file minimal.dnimm3")
		} else if (args.get(0).equals("--help")) {
			print("TODO Help")
			return
		}
		else{
			fname = args.get(0)
		}

		//TODO change DeterministicFunction to UpperCase, change StoppingRules and SimulationScenario
		val dnimm3Uri = URI.createFileURI(fname)
		val omnetUri = URI.createFileURI("omnet_base.omnetmodel") //this is some existing example - ideally should be empty file
		val resourceSet = new ResourceSetImpl()

        // register QPME Model
        val packageRegistry = resourceSet.getPackageRegistry()
        //OMNETPackageImpl.initializePackageContents()
        packageRegistry.put(OMNETPackage.eNS_URI, OMNETPackage.eINSTANCE)
        packageRegistry.put(Dnimm3Package.eNS_URI, Dnimm3Package.eINSTANCE) //dnimm
        
        // Globally register XML resource as UMLResource.Factory.Instance
        val extensionFactoryMap = Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap()
        extensionFactoryMap.put("dnimm3", new XMIResourceFactoryImpl())  //we use standard factories for XMI. This means a file with extension .dnimm3 is an xmi
        extensionFactoryMap.put("omnetmodel", new XMIResourceFactoryImpl()) //the same for omnet //this is not ned! //.omnetmodel -> model-to-text -> ned, ini, rest

        val dniResource = resourceSet.getResource(dnimm3Uri, true)
        val omnetResource = resourceSet.getResource(omnetUri, true)

		do_transformation(dniResource, omnetResource)		
		
		new XMIResourceImpl => [
			contents.add(omnetResource.contents.get(0))
			save(new FileOutputStream("./output_omnet.xmi"), null)
		]
		
		//omnetResource.save(new FileOutputStream("output_omnet.xml", false), null)
	}
	
	private def static do_transformation(Resource dniResource, Resource omnetResource) {
		
		val dniModel = dniResource.contents.get(0) as NetworkInfrastructure
		val omnetModel = omnetResource.contents.get(0) as NedFile
		
		println("== Starting transformation ==")
		val dni = dniModel
		
		val ned = omnetModel
		val ini = omnetModel.config
		val routingxml = omnetModel.config.routingconf
		//these elements are present in the base file
		
		//routingxml.filename = '''example_Ã‚Â«dni.dataCenterNameÃ‚Â».xml'''
		//or
		routingxml.filename = "example_"+ dni.dataCenterName + ".xml"
		
		
		val netWork = fac.createNetwork
		netWork.name = "Submodules"
		createSimController(netWork)
		val global = fac.createLocalType
		global.name = "GlobalVars"
		val iniGroup = fac.createOmnetppIniGroup
		iniGroup.name = "General"
		
		// how to filter and iterate over all model entities
		for(e: dniResource.allContents.toIterable()) {
 			println(e)
		}
		
		//how to filter and iterate over all model entities of type
		for(e: dniResource.allContents.toIterable.filter(typeof(NamedElement))) {
 			println(e.name)
		}
		
		// how select only some parts of the model starting from the model root (for DNI, model root is NetworkInfrastructure
		val dniCommonNodes = dni.structure.nodes.filter[it.sdntype instanceof Common].map[it as Node]
		for(n : dniCommonNodes){
			parseNodes(netWork, iniGroup, n)
		}
		// get communicating Applications 
		val CommunicatingApplications = dni.traffic.software.map[it as CommunicatingApplication]
		for(n : CommunicatingApplications){
			parseCommunicatingApp(netWork, iniGroup, n)
		}
		//get Flows
		val flows = dni.traffic.flows.map[it as Flow]
		for(n : flows){
			parseFlow(netWork, n)
		}
		//get Connnections and links
		val connections = fac.createConnections
		val links = dni.structure.links.map[it as Link]
		for(l : links){
			parseLinks(netWork, connections,l)
			netWork.connections = connections
		}
		//get directsion
		val directions = dni.configuration.routes.directions
		for(d : directions)
		{
			parseDirections(iniGroup, d)
			//Parse Directions [flowName viaInterface Probability]
			//attention multiple directions rules can be linked together in one large string in omnet
			//example [flowName1 viaInterface1 Probability1 flowName2 viaInterface2 Probability2]
		}
		//get routes
		val routes = dni.configuration.routes.flowRoutes
		for (r : routes)
		{
			parseFlowIni(iniGroup, r)
		}
		
		//fac.createModuleType
		/*
		////////////////////////////////////////////
		//example Channel
		val chan = fac.createChannel
		chan.name = "Cat6"
		chan.extends = "ned.DatarateChannel"
			val dP = fac.createDoubleParameter
			dP.name = "datarate"
			dP.unit = "Gbps"
			dP.value = 1.0
		chan.parameters.add(dP)
		ned.definitions.add(chan as Channel)
		
		
		////////////////////////////////////////////
		//example two Nodes as SubModules Type Node
		val node1 = fac.createSubModule
		node1.name = "Node1"
		node1.type = "Node"
		val node2 = fac.createSubModule
		node2.name = "Node2"
		node2.type = "Node"


		////////////////////////////////////////////
		//example connections
		val cons = fac.createConnections
		cons.allowunconnected = true
		val con = fac.createChannelConnectionItem
		con.direction = Direction.LEFTTORIGHT;
		 
		con.channelId = "Cat6"
		val portI = fac.createSimplePort
		portI.name = "in"
		portI.increment = true
		val portO = fac.createSimplePort
		portO.name = "out"
		portO.increment = true
		con.leftPort = portI
		con.rightPort = portO
		cons.connections.add(con)
		*/
		
		
		//add to nedFile
		//compMod.submodules.add(node1)
		//compMod.submodules.add(node2)
		//compMod.connections = cons


		ned.definitions.add(global as LocalType)
		ned.definitions.add(netWork as Network)
		ini.sections.add(iniGroup)
		println("== Transformation finished ==")
	}
	
	//get Nodes and type
	private def static parseNodes(Network net, OmnetppIniGroup inig, Node n)
	{
		val node = fac.createSubModule
		node.name = n.name
		//get all Network Interfaces
		val dniInterfaces = n.interfaces.map[it as NetworkInterface]
		for (iF : dniInterfaces )
		{
			val netIF = fac.createSubModule
			netIF.name = iF.name
			netIF.type = "NetworkInterface"
			net.submodules.add(netIF)
			
			writeIniLine(inig, "General", iF.name + ".mtu", iF.performance.MTU.toString(), "")
			writeIniLine(inig, "General", iF.name + ".isUp", iF.performance.isIsUp.toString(), "")
		}
		//TODO:
		//ADD VALUES TO INI
		//TODO
		// read values correctly and parse them correctly currently 
		// toSTring will return like this dnimm3.impl.ConstantDoubleVariableImpl@80169cf (value: 1.0)
		//CAST TO CONSTANTDOUBLEVARIABLE SEE BELOW
		// need to access value 1.0 directly and use DoubleParameter
		if (n.intermediate != null)
		{
			
			writeIniLine(inig, "General", n.name + ".nativeSwitchingCapacityPPS",
				n.intermediate.performance.switchingCapacityPPS.toString(),
				n.intermediate.performance.switchingCapacityPPS.unit.toString())
			writeIniLine(inig, "General", n.name + ".nativeForwardingLatency",
				n.intermediate.performance.forwardingLatency.toString(),
			    (n.intermediate.performance.forwardingLatency as ConstantDoubleVariable).value.toString)
			writeIniLine(inig, "General", n.name + ".nativeForwardingBandwithBPS",
				n.intermediate.performance.forwardingBandwidthBPS.toString(),
				n.intermediate.performance.forwardingBandwidthBPS.unit.toString())
		}				
		node.type = "Node"
		net.submodules.add(node)
		//TODO:
		//missing if SDN thenn ad sdnFlowRules isDerfault responsedelay etc.
		if (n.sdntype != null)
			println("error SDN-Application missing parse rules ." +n)
	}
	
	//create communication application
	private def static parseCommunicatingApp(Network net, OmnetppIniGroup inig,  CommunicatingApplication a)
	{
		val app = fac.createSubModule
		app.name = a.name
		app.type = "CommunicatingApplication"
		net.submodules.add(app)
		if (a.deployedOn!=null )
		{
			writeIniLine(inig, "General", a.name + ".deployedOnNode", a.deployedOn.node.name, "")
			//create XMLFILE here
			generateTraffic(a)
			}
	}
	
	//create flows
	private def static parseFlow(Network net, Flow f)
	{
		val dir = fac.createSubModule
		dir.name = f.name
		dir.type = "Flow"
		net.submodules.add(dir)
	}

	private def static parseLinks( Network net, Connections cons, Link l)
	{
		val con = fac.createChannelConnectionItem
		
		val interFaces = l.connects
		if (interFaces!=null){
			val portL = fac.createPortID
			portL.name = interFaces.get(0).name+".tx"
			val portR = fac.createPortID
			portR.name = interFaces.get(1).name+".rx"
			con.channelId = l.name
			con.leftPort = portL
			con.rightPort = portR
			
			val perf = fac.createChannel
			perf.extends = "ned.DatarateChannel"
			perf.name = l.name
			
			//TODO read values correctly and parse them correctly currently 
			// toSTring will return dnimm3.impl.ConstantDoubleVariableImpl@80169cf (value: 1.0)
			// need to access value 1.0 directly and use DoubleParameter
			val value = fac.createStringParameter
			value.name = l.name
			value.unit = l.performance.maxSupportedBandwidth.unit.toString();
			value.value = l.performance.maxSupportedBandwidth.toString();

			perf.parameters.add(value)
			net.localTypes.add(perf)
			}
			
		cons.allowunconnected = true
		cons.connections.add(con)
	}
	
	//create Simulation Controller application
	private def static createSimController(Network net)
	{
		val simCont = fac.createSubModule
		simCont.name = "SimController"
		simCont.type = "SimulationController"
		net.submodules.add(simCont)
	}
	
	//Write into IniGroup contains omnetpp.ini settings
	private def static writeIniLine(OmnetppIniGroup ini, String nodeSelector, String parameter, String value, String prefix)
	{
		var iniLine = fac.createOmnetppIniLine
		iniLine.value = value
		iniLine.nodeSelector = nodeSelector
		iniLine.parameter = parameter
		iniLine.prefix = prefix
		ini.lines.add(iniLine)
	}
	
	//get Directions
	private def static parseDirections(OmnetppIniGroup ini, dnimm3.Direction d)
	{
		writeIniLine(ini, "General", d.onNode.description, d.flow.name + " " + d.via.name + " " + d.probability.toString(), "name: " + d.name )
	}
	
	//get Flowsettings for Ini
	private def static parseFlowIni(OmnetppIniGroup ini, FlowRoute r)
	{
		writeIniLine(ini,"General", r.flow.name + ".SourceSofware", r.start.node.name, "")
		writeIniLine(ini,"General", r.flow.name + ".DestinationSofware", r.end.node.name, "")
		writeIniLine(ini,"General", r.flow.name + ".sendInterval", "0", "s")

		//TODO:
		//get DataSize
	}
	
	private def static generateTraffic(CommunicatingApplication comApp)
	{
		//Create xml trafic file
		//Example:
	/* 
	 * <?xml version="1.0" encoding="UTF-8" standalone="no"?>
	 * 		<workload>
	 * 			<actions id="0">
	 * 				<action id="0" transmit="Flow2" waitTime="0.0"/>
	 * 				<action id="1" transmit="Flow2" waitTime="0.0"/>
	 *			</actions>
	 * 		</workload>
	 * */
		
	}
	

}
